/* Copyright 2016 Alex 'AdUser' Z (ad_user@runbox.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include "source.h"

#include <limits.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <glob.h>
#include <time.h>

#define MODNAME "files"

#define FREE(x)   { free(x);   x = NULL; }
#define FCLOSE(x) { if (x) fclose(x); x = NULL; }

typedef struct f2b_file_t {
  struct f2b_file_t *next;
  uint32_t lines;
  uint32_t stag;
  struct stat st;
  FILE *fd;
  char path[];
} f2b_file_t;

struct _config {
  void (*logcb)(enum loglevel lvl, const char *msg);
  f2b_file_t *files;
  f2b_file_t *current;
  int flags;
  bool missingok;
  char path[];
};

#include "source.c"

static bool
file_open(const cfg_t *cfg, f2b_file_t *file) {
  FILE *fd;
  struct stat st;

  assert(file != NULL);

  if (stat(file->path, &st) != 0)
    return false;

  if (!(S_ISREG(st.st_mode) || S_ISFIFO(st.st_mode))) {
    log_msg(cfg, error, "not a regular file/fifo: %s", file->path);
    return false;
  }

  if ((fd = fopen(file->path, "r")) == NULL) {
    log_msg(cfg, error, "can't open file: %s -- %s", file->path, strerror(errno));
    return false;
  }

  if (S_ISREG(st.st_mode) && fseek(fd, 0, SEEK_END) < 0) {
    log_msg(cfg, error, "fseek error: %s -- %s", file->path, strerror(errno));
    FCLOSE(fd);
    return false;
  }

  memcpy(&file->st, &st, sizeof(st));
  file->fd = fd;
  file->stag = fnv_32a_str(file->path, FNV1_32A_INIT);

  return true;
}

static void
file_close(f2b_file_t *file) {
  assert(file != NULL);

  FCLOSE(file->fd);
  file->lines = 0;
  memset(&file->st, 0, sizeof(struct stat));
}

static bool
file_rotated(const cfg_t *cfg, f2b_file_t *file) {
  struct stat st;
  assert(file != NULL);

  if (file->fd == NULL)
    return true;

  if (stat(file->path, &st) != 0) {
    log_msg(cfg, error, "stat(): %s -- %s", file->path, strerror(errno));
    return true;
  }

  if (file->st.st_dev  != st.st_dev ||
      file->st.st_ino  != st.st_ino ||
      file->st.st_size  > st.st_size) {
    log_msg(cfg, info, "file replaced: %s", file->path);
    return true;
  } else if (file->st.st_size < st.st_size) {
    memcpy(&file->st, &st, sizeof(struct stat));
  }

  return false;
}

static bool
file_getline(const cfg_t *cfg, f2b_file_t *file, char *buf, size_t bufsize) {
  char *p;
  assert(file != NULL);
  assert(buf != NULL);

  /* fread()+EOF set is implementation defined */
  if (fgets(buf, bufsize, file->fd) != NULL) {
    if ((p = strchr(buf, '\n')) != NULL)
      *p = '\0'; /* strip newline(s) */
    file->lines++;
    return true;
  }
  if (feof(file->fd))
    clearerr(file->fd);
  if (ferror(file->fd)) {
    log_msg(cfg, error, "file error: %s -- %s", file->path, strerror(errno));
    clearerr(file->fd);
  }

  return false;
}

cfg_t *
create(const char *init) {
  cfg_t *cfg = NULL;

  if (init == NULL) return NULL;

  size_t ilen = strlen(init) + 1;
  if (ilen <= 1) return NULL;

  if ((cfg = calloc(1, sizeof(cfg_t) + ilen)) == NULL)
    return NULL;
  cfg->flags |= MOD_TYPE_SOURCE;
  cfg->flags |= MOD_NEED_FILTER;

  cfg->logcb = &logcb_stub;
  strlcpy(cfg->path, init, ilen);
  cfg->flags |= MOD_IS_READY;

  return cfg;
}

bool
config(cfg_t *cfg, const char *key, const char *value) {
  assert(cfg != NULL);
  assert(key   != NULL);
  assert(value != NULL);

  if (strcmp(key, "missingok") == 0) {
    cfg->missingok = (strcmp(value, "yes") == 0) ? true : false;
    return true;
  }

  return false;
}

bool
start(cfg_t *cfg) {
  f2b_file_t *file = NULL;
  int glob_flags = GLOB_MARK | GLOB_NOESCAPE;
  glob_t globbuf;

  assert(cfg != NULL);

  if (cfg->missingok)
    glob_flags |= GLOB_NOCHECK; /* allow no matches */

  if (glob(cfg->path, glob_flags, NULL, &globbuf) != 0)
    return NULL;

  for (size_t i = 0; i < globbuf.gl_pathc; i++) {
    size_t slen = strlen(globbuf.gl_pathv[i]) + 1;
    if ((file = calloc(1, sizeof(f2b_file_t) + slen)) == NULL) {
      log_msg(cfg, error, "can't allocate memory: %s", strerror(errno));
      continue;
    }
    strlcpy(file->path, globbuf.gl_pathv[i], slen);
    if (!file_open(cfg, file) && !cfg->missingok) {
      FREE(file);
      continue;
    }
    if (cfg->files) {
      file->next = cfg->files;
      cfg->files = file;
    } else {
      cfg->files = file;
    }
  }

  globfree(&globbuf);
  return true;
}

bool
stop(cfg_t *cfg) {
  f2b_file_t *next = NULL;

  assert(cfg != NULL);

  for (; cfg->files != NULL; cfg->files = next) {
    next = cfg->files->next;
    file_close(cfg->files);
    FREE(cfg->files);
  }

  return true;
}

uint32_t
next(cfg_t *cfg, char *buf, size_t bufsize, bool reset) {
  assert(cfg != NULL);
  assert(buf != NULL);
  assert(bufsize > 0);

  if (reset || cfg->current == NULL)
    cfg->current = cfg->files;

  for (f2b_file_t *file = cfg->current; file != NULL; file = file->next) {
    if (file->fd != NULL && file_rotated(cfg, file))
      file_close(file);
    if (file->fd == NULL && !file_open(cfg, file))
      continue;
    if (file_getline(cfg, file, buf, bufsize))
      return file->stag;
  }

  return 0;
}

bool
stats(cfg_t *cfg, char *buf, size_t bufsize) {
  char tmp[PATH_MAX + 512];
  time_t mtime;
  const char *fmt =
    "- path: %s\n"
    "  info: fd=%d inode=%d size=%ld pos=%ld tag=%08X readl=%lu mtime=%lu\n";
  int fd, ino; off_t sz; long pos;
  assert(cfg != NULL);
  assert(buf != NULL);
  assert(bufsize > 0);

  if (buf == NULL || bufsize == 0)
    return false;

  for (f2b_file_t *f = cfg->files; f != NULL; f = f->next) {
    if (f->fd) {
      fd = fileno(f->fd), ino = f->st.st_ino, sz = f->st.st_size, pos = ftell(f->fd);
      mtime = f->st.st_mtime;
    } else {
      fd = -1, ino = -1, sz = 0, pos = -1, mtime = 0;
    }
    snprintf(tmp, sizeof(tmp), fmt, f->path, fd, ino, sz, pos, f->stag, f->lines, mtime);
    strlcat(buf, tmp, bufsize);
  }

  return true;
}

void
destroy(cfg_t *cfg) {
  assert(cfg != NULL);

  FREE(cfg);
}
